/*
 * Copyright (c) 2019 QNX Software Systems. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <j7_cpsw.h>
#include <io-pkt/iopkt_driver.h>
#include <sys/io-pkt.h>
#include <sys/syspage.h>
#include <sys/device.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <nw_thread.h>
#include <sys/netmgr.h>
#include <net/ifdrvcom.h>
#include <sys/sockio.h>
#include <device_qnx.h>
#include <net/netbyte.h>
#include <quiesce.h>

#include "cpsw_lld_if.h"

#define NBPFILTER 1
#if NBPFILTER > 0
#include <net/bpf.h>
#include <net/bpfdesc.h>
#endif


struct _iopkt_drvr_entry IOPKT_DRVR_ENTRY_SYM(cpsw) = IOPKT_DRVR_ENTRY_SYM_INIT(cpsw_entry);

#ifdef VARIANT_a
#include <nw_dl.h>
/* This is what gets specified in the stack's dl.c */
struct nw_dll_syms cpsw_syms[] = {
        {"iopkt_drvr_entry", &IOPKT_DRVR_ENTRY_SYM(cpsw)},
        {NULL, NULL}
};
#endif

int cpsw_attach(struct device *, struct device *, void *);
int cpsw_detach(struct device *, int);

CFATTACH_DECL(cpsw,
    sizeof(struct cpsw_dev),
    NULL,
    cpsw_attach,
    cpsw_detach,
    NULL);

attach_args_t   attach_args;
int single      = 1; /* single instance of port*/
struct cpsw_dev *g_cpsw = NULL;

/* Driver options */
static char *cpsw_opts[] = {
    "verbose",
#define CPSWOPT_VERBOSE  0
    "mac-to-mac",
#define CPSWOPT_MACTOMAC 1
    "speed",
#define CPSWOPT_SPEED    2
    "p0mac",
#define CPSWOPT_P0MAC    3
    "ptp",
#define CPSWOPT_PTP      4
        NULL
};

#define CPSW_TX_FREEQ_THRESHOLD 64


/*****************************************************************************/
/*  RX thread quiesce fuction                                                */
/*****************************************************************************/
void cpsw_rx_thread_quiesce (void *arg, int die)
{
    struct cpsw_dev     *cpsw = arg;

    MsgSendPulse(cpsw->coid, SIGEV_PULSE_PRIO_INHERIT,
        CPSW_QUIESCE_PULSE, die);

    return;
}

/*****************************************************************************/
/*  RX thread init fuction                                                   */
/*****************************************************************************/
static int cpsw_rx_thread_init (void *arg)
{
    struct cpsw_dev     *cpsw = arg;
    struct nw_work_thread   *wtp = WTP;

#if defined (CPSW2G)
    pthread_setname_np(gettid(), "cpsw2g Rx");
#elif defined (CPSW9G)
    pthread_setname_np(gettid(), "cpsw9g Rx");
#else
#error "Unsupported!"
#endif

    wtp->quiesce_callout = cpsw_rx_thread_quiesce;
    wtp->quiesce_arg = cpsw;
    return EOK;
}


/*****************************************************************************/
/* Parse the options                                                         */
/*****************************************************************************/
void cpsw_parse_options(struct cpsw_dev *cpsw, const char *optstring,
                          nic_config_t *cfg)

{
    char    *value;
    int     opt;
    char    *options, *freeptr;
    int     tmp;

    if (optstring == NULL) {
        return;
    }

    /* getsubopt() is destructive */
    freeptr = options = strdup (optstring);

    while (options && *options != '\0') {
    opt = getsubopt (&options, cpsw_opts, &value);

    switch (opt) {
        case CPSWOPT_VERBOSE:
            if (cpsw != NULL) {
                if (value) {
                    cpsw->cfg.verbose = strtoul(value, 0, 0);
                    slogf(_SLOGC_NETWORK, _SLOG_ERROR, "%s:%d Verbose -->%d",
                        __FUNCTION__, __LINE__, cpsw->cfg.verbose);
                } else {
                    cpsw->cfg.verbose++;
                }
            }
            break;
        case CPSWOPT_MACTOMAC:
            if (cpsw != NULL) {
                tmp = strtoul(value, 0, 0);
                if ((tmp != 0) && (tmp != 1)) {
                    slogf(_SLOGC_NETWORK, _SLOG_ERROR,
                      "%s:%d Error: Invalid mac-to-mac value %s", __FUNCTION__, __LINE__, value);
                } else {
                    cpsw->mac_to_mac = tmp;
                    slogf(_SLOGC_NETWORK, _SLOG_INFO, "%s:%d mac_to_mac -->%d",
                        __FUNCTION__, __LINE__, cpsw->mac_to_mac);
                }
            }
            break;
        case CPSWOPT_SPEED:
            if (cpsw != NULL) {
                tmp = strtoul(value, 0, 0);
                if ((tmp != 100) && (tmp != 1000)) {
                    slogf(_SLOGC_NETWORK, _SLOG_ERROR,
                      "%s:%d Error: Invalid Port 0 speed %s", __FUNCTION__, __LINE__, value);
                } else {
                    cpsw->speed |= tmp;
                    slogf(_SLOGC_NETWORK, _SLOG_INFO, "%s:%d speed -->%d",
                        __FUNCTION__, __LINE__, cpsw->speed);
                }
            }
            break;
        case CPSWOPT_P0MAC:
            if (cpsw != NULL) {
                if (nic_strtomac(value, cfg->current_address)) {
                    slogf(_SLOGC_NETWORK, _SLOG_WARNING,
                        "%s:%d Error: Invalid mac address %s", __FUNCTION__, __LINE__, value);
                }
            }
            break;
        case CPSWOPT_PTP:
            if (cpsw != NULL) {
                cpsw->ptp_enable = strtoul(value, 0, 0);
            }
            break;
        default:
            slogf(_SLOGC_NETWORK, _SLOG_WARNING,
              "%s:%d Error: Skipping unknown option %s", __FUNCTION__, __LINE__, value);
            break;
        }
    }

    (free)(freeptr);
    return;
}


/*****************************************************************************/
/*  Initial driver entry point.                                              */
/*****************************************************************************/
int
cpsw_entry(void *dll_hdl,  struct _iopkt_self *iopkt, char *options)
{
    int     instance;
    struct device   *dev;
    struct drvcom_config    *dcon;
    struct ifnet    *ifp;
    int err;

    slogf(_SLOGC_NETWORK, _SLOG_ERROR, "%s:%d Entry -->", __FUNCTION__, __LINE__);

    /* Check if it is already mounted by doing a "nicinfo" on each interface */
    dcon = (malloc)(sizeof(*dcon));
    if (dcon == NULL) {
        slogf(_SLOGC_NETWORK, _SLOG_ERROR,
              "%s:%d Error: malloc failed", __FUNCTION__, __LINE__);
        return ENOMEM;
    }
    IFNET_FOREACH(ifp) {
        dcon->dcom_cmd.ifdc_cmd = DRVCOM_CONFIG;
        dcon->dcom_cmd.ifdc_len = sizeof(dcon->dcom_config);
        err = ifp->if_ioctl(ifp, SIOCGDRVCOM, (caddr_t)dcon);
        if ((err == EOK) && (dcon->dcom_config.num_io_windows > 0) &&
            (dcon->dcom_config.io_window_base[0] == J721E_CPSW_BASE)) {
            slogf(_SLOGC_NETWORK, _SLOG_ERROR,
                  "%s:%d Error: Driver already loaded for %s",
                  __FUNCTION__, __LINE__, ifp->if_xname);
            return EBUSY;
        }
    }
    (free)(dcon);

    /* initialize to whatever you want to pass to cpsw_attach() */
    memset(&attach_args, 0x00, sizeof(attach_args));

    attach_args.iopkt = iopkt;
    attach_args.options = options;

    /* parse options */

    attach_args.cfg.num_io_windows = 1;
    attach_args.cfg.io_window_base[0] = J721E_CPSW_BASE;
    attach_args.cfg.io_window_size[0] = J721E_CPSW_SIZE;

    attach_args.cfg.num_irqs = 0;

    /* we have single instance */
    single = 1;
    attach_args.cfg.device_index = instance = 0;

    for (instance = 0;;) {
        /* Apply detection criteria */

        /* Found one */
        dev = NULL; /* No Parent */
#if defined (CPSW2G)
        if (dev_attach("am", options, &cpsw_ca, &attach_args,
#elif defined (CPSW9G)
        if (dev_attach("an", options, &cpsw_ca, &attach_args,
#endif
            &single, &dev, NULL) != EOK) {
            break;
        }
        dev->dv_dll_hdl = dll_hdl;
        instance++;


        if (/* done_detection || */ single)
            break;
    }

    if (instance > 0)
        return EOK;

    return ENODEV;
}

/*****************************************************************************/
/*                                                                           */
/*****************************************************************************/
int
cpsw_attach(struct device *parent, struct device *self, void *aux)
{
    struct cpsw_dev     *cpsw;
    struct ifnet        *ifp;
    struct nw_stk_ctl   *sctlp;
    char                                *options;
    uint8_t         enaddr[ETHER_ADDR_LEN];

    slogf(_SLOGC_NETWORK, _SLOG_ERROR, "%s:%d Entry -->", __FUNCTION__, __LINE__);

    /* initialization and attach */
    options = attach_args.options;
    cpsw = (struct cpsw_dev *)self;
    ifp = &cpsw->sc_ec.ec_if;
    ifp->if_softc = cpsw;
    cpsw->sc_iopkt = iopkt_selfp;
    cpsw->sc_iid = -1; /* not attached yet */

    sctlp = &stk_ctl;

    /* Copy various config parameters initilized in cpsw_entry */
    memcpy(&cpsw->cfg, &attach_args.cfg, sizeof(nic_config_t));

    /* Parse options */
    cpsw_parse_options(cpsw, options, &cpsw->cfg);

    /* Set interface name */
    strcpy (ifp->if_xname, cpsw->sc_dev.dv_xname);
#if defined (CPSW2G)
    strcpy ((char *) cpsw->cfg.uptype, "am");
    strcpy ((char *) cpsw->cfg.device_description, "cpsw2g");
#elif defined (CPSW9G)
    strcpy ((char *) cpsw->cfg.uptype, "an");
    strcpy ((char *) cpsw->cfg.device_description, "cpsw9g");
#endif

    /* Ethernet stats we are interested in */
    cpsw->stats.un.estats.valid_stats =
        NIC_ETHER_STAT_INTERNAL_TX_ERRORS |
        NIC_ETHER_STAT_INTERNAL_RX_ERRORS |
        NIC_ETHER_STAT_NO_CARRIER |
        NIC_ETHER_STAT_XCOLL_ABORTED |
        NIC_ETHER_STAT_SINGLE_COLLISIONS |
        NIC_ETHER_STAT_MULTI_COLLISIONS |
        NIC_ETHER_STAT_LATE_COLLISIONS |
        NIC_ETHER_STAT_TX_DEFERRED |
        NIC_ETHER_STAT_ALIGN_ERRORS |
        NIC_ETHER_STAT_FCS_ERRORS;

    /* Generic networking stats we are interested in */
    cpsw->stats.valid_stats =
        NIC_STAT_TX_FAILED_ALLOCS |
        NIC_STAT_RX_FAILED_ALLOCS |
        NIC_STAT_RXED_MULTICAST |
        NIC_STAT_RXED_BROADCAST |
        NIC_STAT_TXED_BROADCAST |
        NIC_STAT_TXED_MULTICAST;

    cpsw->cfg.priority = IRUPT_PRIO_DEFAULT;
    cpsw->cfg.lan = cpsw->sc_dev.dv_unit;
    cpsw->cfg.media_rate = -1;
    cpsw->cfg.duplex = -1;
    cpsw->force_link = -1;
    cpsw->cfg.flags |= NIC_FLAG_LINK_DOWN;
    cpsw->linkup = 0;
    if  (!cpsw->mac_to_mac) {
        /* default to 1Gbps if not mac-to-mac */
        cpsw->speed = 1000;
    }

    cpsw->cfg.mtu = ETHERMTU;
    cpsw->cfg.mru = ETHERMTU;
    cpsw->cfg.flags |= NIC_FLAG_MULTICAST;
    cpsw->cfg.mac_length = ETHER_ADDR_LEN;

    /* set capabilities */
    ifp->if_flags = IFF_BROADCAST | IFF_SIMPLEX | IFF_MULTICAST;
    cpsw->sc_ec.ec_capabilities |= ETHERCAP_VLAN_MTU;

    /* Set callouts */
    ifp->if_ioctl = cpsw_ioctl;
    ifp->if_start = cpsw_start;
    ifp->if_init = cpsw_init;
    ifp->if_stop = cpsw_stop;
    IFQ_SET_READY(&ifp->if_snd);

    /* hook up so media devctls work */
    bsd_mii_initmedia (cpsw);

    /* some IO registers are in privileged memory */
    if (ThreadCtl(_NTO_TCTL_IO_PRIV, NULL) == -1) {
        slogf(_SLOGC_NETWORK, _SLOG_ERROR, "%s:%d Error: ThreadCtl failed: %s",
                __FUNCTION__, __LINE__, strerror(errno));
        /* Continue in user mode  */
    }

    /* If media was not specified on cmdline, default to NIC_MEDIA_802_3 */
    if (cpsw->cfg.media == -1) {
        cpsw->cfg.media = NIC_MEDIA_802_3;
        cpsw->stats.media = cpsw->cfg.media;
    }
    else {
        cpsw->stats.media = cpsw->cfg.media;
    }

    if (cpsw->cfg.mtu == 0 || cpsw->cfg.mtu > (ETHER_MAX_LEN - ETHER_CRC_LEN)) {
        cpsw->cfg.mtu = (ETHER_MAX_LEN - ETHER_CRC_LEN);
    }

    if (cpsw->cfg.mru == 0 || cpsw->cfg.mru > (ETHER_MAX_LEN - ETHER_CRC_LEN)) {
        cpsw->cfg.mru = (ETHER_MAX_LEN - ETHER_CRC_LEN);
    }

    callout_init(&cpsw->phy_callout);

    cpsw->chid = ChannelCreate(0);
    cpsw->coid = ConnectAttach(ND_LOCAL_NODE, 0, cpsw->chid, _NTO_SIDE_CHANNEL, 0);

    SIGEV_PULSE_INIT(&cpsw->isr_event, cpsw->coid, sctlp->rx_prio, CPSW_RX_PULSE, 0);

    if (nw_pthread_create(&cpsw->tid, NULL,
          cpsw_rx_thread, cpsw, 0,
          cpsw_rx_thread_init, cpsw) != EOK) {
        slogf(_SLOGC_NETWORK, _SLOG_ERROR, "%s:%d Error: nw_pthread_create failed: %s",
                __FUNCTION__, __LINE__, strerror(errno));
        return ENOMEM;
    }

    CpswApp_InitObj();
    CpswApp_GetMacAddr(enaddr);
    memcpy(cpsw->cfg.permanent_address, enaddr, ETHER_ADDR_LEN);
    /* Not specified on command line, copy to current */
    if (!memcmp(cpsw->cfg.current_address, "\0\0\0\0\0\0", ETHER_ADDR_LEN)) {
        memcpy(cpsw->cfg.current_address, cpsw->cfg.permanent_address, ETHER_ADDR_LEN);
    }

    if_attach(ifp);
    ether_ifattach(ifp, cpsw->cfg.current_address);

    if (cpsw->cfg.verbose) {
        nic_dump_config (&cpsw->cfg);
    }

    cpsw->sc_sdhook = shutdownhook_establish(cpsw_shutdown, cpsw);

    // for any local use
    g_cpsw = cpsw;
    return EOK;
}


/*****************************************************************************/
/*                                                                           */
/*****************************************************************************/
int
cpsw_init(struct ifnet *ifp)
{
    struct cpsw_dev *cpsw;
    struct nw_work_thread       *wtp = WTP;

    cpsw = ifp->if_softc;

    if (cpsw->dying == 1)
        return (0);

    if (cpsw->cfg.verbose & DEBUG_TRACE) {
        slogf(_SLOGC_NETWORK, _SLOG_ERROR, "%s:%d Entry --> ",
            __FUNCTION__, __LINE__);
    }

    /*
     * - enable hardware.
     *   - look at ifp->if_capenable_[rx/tx]
     *   - enable promiscuous / multicast filter.
     * - attach to interrupt.
     */

    if(memcmp(cpsw->cfg.current_address, LLADDR(ifp->if_sadl), ifp->if_addrlen)) {
        memcpy(cpsw->cfg.current_address, LLADDR(ifp->if_sadl), ifp->if_addrlen);
        /* update the hardware */
    }

    if ((ifp->if_flags & IFF_RUNNING) == 0) {
        CpswApp_Init(cpsw->mac_to_mac, cpsw->speed, cpsw->cfg.current_address);
#if defined (CPSW2G)
        CpswApp_InitPhy();
        if (!cpsw->mac_to_mac) {
            CpswApp_WaitTillPortUp();
        }
        if( cpsw->ptp_enable)
        {
            CpswApp_PtpInit();
        }
#elif defined (CPSW9G)
       /* Not required for cpsw9g */
#endif
        callout_msec (&cpsw->phy_callout, PHY_CALLOUT_DELAY,
                          cpsw_MonitorPhy, cpsw);
        if  (cpsw->mac_to_mac) {
            cpsw->cfg.connector = NIC_CONNECTOR_UNKNOWN;
        }
        else {
            cpsw->cfg.connector = NIC_CONNECTOR_MII;
        }
        cpsw->cfg.phy_addr = 0;
        cpsw->cfg.duplex = 1;
        if (cpsw->speed == 100) {
            cpsw->cfg.media_rate = 100 * 1000L;
        }
        else {
            /* default to 1Gbps */
            cpsw->cfg.media_rate = 1000 * 1000L;
        }
        cpsw->cfg.flags &= ~NIC_FLAG_LINK_DOWN;
        cpsw->linkup = 1;
        if_link_state_change(ifp, LINK_STATE_UP);
    }

    NW_SIGLOCK_P (&ifp->if_snd_ex, cpsw->sc_iopkt, wtp);
    ifp->if_flags_tx |= IFF_RUNNING;
    ifp->if_flags_tx &= ~IFF_OACTIVE;
    NW_SIGUNLOCK_P (&ifp->if_snd_ex, cpsw->sc_iopkt, wtp);

    ifp->if_flags |= IFF_RUNNING;

    if (cpsw->cfg.verbose & DEBUG_TRACE) {
        slogf(_SLOGC_NETWORK, _SLOG_ERROR, "%s:%d <--Exit ",
            __FUNCTION__, __LINE__);
    }
    return EOK;
}

void
cpsw_stop(struct ifnet *ifp, int disable)
{
    struct cpsw_dev *cpsw = ifp->if_softc;
    struct nw_work_thread   *wtp = WTP;
    struct _iopkt_self  *iopkt = cpsw->sc_iopkt;

    if (cpsw->cfg.verbose & DEBUG_TRACE) {
        slogf(_SLOGC_NETWORK, _SLOG_ERROR, "%s:%d Entry -->",
            __FUNCTION__, __LINE__);
    }

    /* Lock out the transmit side */
    NW_SIGLOCK_P (&ifp->if_snd_ex, iopkt, wtp);

    ifp->if_flags_tx |= IFF_OACTIVE;

    /* Release any queued transmit buffers */

    ifp->if_flags_tx &= ~(IFF_RUNNING | IFF_OACTIVE);
    /* Done with the transmit side */
    NW_SIGUNLOCK_P (&ifp->if_snd_ex, iopkt, wtp);


    /*
     * - Cancel any pending io
     * - Clear any interrupt source registers
     * - Clear any interrupt pending registers
     * - Release any queued transmit buffers.
     */

    if (disable) {
        if (cpsw->linkup) {
            /* Stop monitoring */
            callout_stop(&cpsw->phy_callout);
            CpswApp_Close();
            cpsw->cfg.flags |= NIC_FLAG_LINK_DOWN;
            cpsw->linkup = 0;
            if_link_state_change(ifp, LINK_STATE_DOWN);
        }
    }

    ifp->if_flags &= ~(IFF_RUNNING | IFF_OACTIVE);

    if (cpsw->cfg.verbose & DEBUG_TRACE) {
        slogf(_SLOGC_NETWORK, _SLOG_ERROR, "%s:%d <--Exit",
            __FUNCTION__, __LINE__);
    }
}

/*****************************************************************************/
/*                                                                           */
/*****************************************************************************/

static struct mbuf *cpsw_defrag (struct mbuf *m)

{
    struct mbuf *m2;

    MGET (m2, M_DONTWAIT, MT_DATA);
    if (m2 == NULL) {
        m_freem (m);
        return (NULL);
    }

    M_COPY_PKTHDR (m2, m);

    MCLGET (m2, M_DONTWAIT);
    if ((m2->m_flags & M_EXT) == 0) {
        m_freem (m);
        m_freem (m2);
        return (NULL);
    }

    /* Paranoid ? */
    if (m->m_pkthdr.len > m2->m_ext.ext_size) {
        m_freem (m);
        m_freem (m2);
        return (NULL);
    }

    m_copydata (m, 0, m->m_pkthdr.len, mtod(m2, caddr_t));
    m2->m_pkthdr.len = m2->m_len = m->m_pkthdr.len;

    m_freem(m);

    return (m2);
}


/*****************************************************************************/
/*                                                                           */
/*****************************************************************************/
void
cpsw_start(struct ifnet *ifp)
{
    struct cpsw_dev     *cpsw = ifp->if_softc;
    struct mbuf     *m;
    struct mbuf *m2;
    struct nw_work_thread   *wtp = WTP;
    int num_frags;
    uint32_t num_free = 0;


    if (cpsw->cfg.verbose & DEBUG_TRACE) {
        slogf(_SLOGC_NETWORK, _SLOG_ERROR, "%s:%d Entry -->", __FUNCTION__, __LINE__);
    }

    if ((ifp->if_flags_tx & IFF_RUNNING) == 0) {
        NW_SIGUNLOCK_P (&ifp->if_snd_ex, cpsw->sc_iopkt, wtp);
        return;
    }

    ifp->if_flags_tx |= IFF_OACTIVE;

    for (;;) {
        /* Are there buffers available to send this packet */
        num_free = CpswApp_GetTxFreeQCnt();
        if(num_free <= CPSW_TX_FREEQ_THRESHOLD)
        {
            CpswApp_retrieveFreeTxPkts();
        }

        num_free = CpswApp_GetTxFreeQCnt();
        if(num_free == 0) {
            slogf(_SLOGC_NETWORK, _SLOG_WARNING, "%s:%d No buffers available for tx ", __FUNCTION__, __LINE__);
            NW_SIGUNLOCK_P (&ifp->if_snd_ex, cpsw->sc_iopkt, wtp);
            return;
        }

        IFQ_DEQUEUE(&ifp->if_snd, m);
        if (m == NULL) {
            break;
        }

        /* m could be multiple buffers, defrag to single */
        m2 = cpsw_defrag(m);
        /* Re-adjust the number of fragments and total length */
        m = m2;
        for (num_frags = 0, m2 = m; m2; num_frags++) {
            m2 = m2->m_next;
        }

        if (cpsw->cfg.verbose & DEBUG_START_PACKET) {
            slogf(_SLOGC_NETWORK, _SLOG_ERROR, "%s:%d packet length %d, protocol is %d txFreeCnt %d",
                __FUNCTION__, __LINE__, m->m_len, m->m_data[23], num_free);
        }

#if defined (CPSW2G)
        /* Nothing to do for CPSW2G */
#elif defined (CPSW9G)
        /* Snoop send packet to find ARP request for IPv4 address */
        if(cpsw->ipaddr[0] == 0)
        {
            if ((m->m_data[12] == 0x08)&&(m->m_data[13] == 0x06)&&
                    (m->m_data[16] == 0x08)&&(m->m_data[17] == 0x00)) {
                cpsw->ipaddr[0] = m->m_data[28];
                cpsw->ipaddr[1] = m->m_data[29];
                cpsw->ipaddr[2] = m->m_data[30];
                cpsw->ipaddr[3] = m->m_data[31];
                slogf(_SLOGC_NETWORK, _SLOG_ERROR, "IPv4 address: %d:%d:%d:%d",
                        cpsw->ipaddr[0], cpsw->ipaddr[1], cpsw->ipaddr[2], cpsw->ipaddr[3]);
                CpswApp_RegisterIPv4Address(cpsw->ipaddr);
            }
        }
#endif

        /* You're now committed to transmitting it */
        CpswApp_SendPkt(m);
        if (cpsw->cfg.verbose & DEBUG_BUFFER) {
            slogf(_SLOGC_NETWORK, _SLOG_ERROR, "%s:%d tx size:%d", __FUNCTION__, __LINE__, m->m_len);
        }

#if NBPFILTER > 0
        /* Pass the packet to any BPF listeners */
        if (ifp->if_bpf) {
            bpf_mtap (ifp->if_bpf, m);
        }
#endif

        ifp->if_opackets++;  // for ifconfig -v
        // or if error:  ifp->if_oerrors++;
    }


    ifp->if_flags_tx &= ~IFF_OACTIVE;
    NW_SIGUNLOCK_P(&ifp->if_snd_ex, cpsw->sc_iopkt, wtp);
}


/*****************************************************************************/
/*                                                                           */
/*****************************************************************************/
int
cpsw_detach(struct device *dev, int flags)
{
    struct cpsw_dev *cpsw;
    struct ifnet    *ifp;

    cpsw = (struct cpsw_dev *)dev;

    if (cpsw->cfg.verbose & DEBUG_TRACE) {
        slogf(_SLOGC_NETWORK, _SLOG_ERROR, "%s:%d Entry -->", __FUNCTION__, __LINE__);
    }

    /*
     * Clean up everything.
     *
     * The interface is going away but io-pkt is staying up.
     */
    ifp = &cpsw->sc_ec.ec_if;

    /* Don't init() while we're dying. */
    cpsw->dying = 1;

    cpsw_stop(ifp, 1);

    ether_ifdetach(ifp);

    if_detach(ifp);

    shutdownhook_disestablish(cpsw->sc_sdhook);

    nw_pthread_reap(cpsw->tid);
    ConnectDetach(cpsw->coid);
    ChannelDestroy(cpsw->chid);

    if (cpsw->cfg.verbose & DEBUG_TRACE) {
        slogf(_SLOGC_NETWORK, _SLOG_ERROR, "%s:%d <--Exit", __FUNCTION__, __LINE__);
    }

    return EOK;
}

/*****************************************************************************/
/*                                                                           */
/*****************************************************************************/
void
cpsw_shutdown(void *arg)
{
    struct cpsw_dev *cpsw;

    /* All of io-pkt is going away.  Just quiet hardware. */

    cpsw = arg;

    if (cpsw->cfg.verbose & DEBUG_TRACE) {
        slogf(_SLOGC_NETWORK, _SLOG_ERROR, "%s:%d Entry -->", __FUNCTION__, __LINE__);
    }

    cpsw_stop(&cpsw->sc_ec.ec_if, 1);

    if (cpsw->cfg.verbose & DEBUG_TRACE) {
        slogf(_SLOGC_NETWORK, _SLOG_ERROR, "%s:%d <--Exit", __FUNCTION__, __LINE__);
    }
}

/*****************************************************************************/
/*                                                                           */
/*****************************************************************************/
void cpsw_MonitorPhy (void *arg)
{
    struct cpsw_dev *cpsw = arg;

#if defined (CPSW2G)
    if (!cpsw->mac_to_mac) {
        CpswApp_Tick();
    }
#elif defined (CPSW9G)
    /* Not required for cpsw9g */
#endif

    callout_msec(&cpsw->phy_callout, PHY_CALLOUT_DELAY,
                 cpsw_MonitorPhy, cpsw);
}


/*****************************************************************************/
/*                                                                           */
/*****************************************************************************/
void SetRxPulse()
{
    MsgSendPulse(g_cpsw->coid, SIGEV_PULSE_PRIO_INHERIT, CPSW_RX_PULSE, 0);
}

/*****************************************************************************/
/*                                                                           */
/*****************************************************************************/
struct mbuf * cpsw_process_rx(void *arg, struct mbuf *rpkt)
{
    struct cpsw_dev     *cpsw = arg;
    struct mbuf         *m;
    struct ifnet        *ifp = &cpsw->sc_ec.ec_if;
    struct nw_work_thread   *wtp = WTP;

    rpkt->m_pkthdr.rcvif = ifp;

#if NBPFILTER > 0
    /* Pass the packet to any BPF listeners */
    if (ifp->if_bpf) {
        bpf_mtap (ifp->if_bpf, rpkt);
    }
#endif

    ifp->if_ipackets++; // for ifconfig -v
    (*ifp->if_input)(ifp, rpkt);

    /* Get a packet/buffer to replace the one that was filled */
    m = m_getcl_wtp(M_DONTWAIT, MT_DATA, M_PKTHDR, wtp);
    if (!m) {
        slogf(_SLOGC_NETWORK, _SLOG_ERROR, "%s:%d Error: m_getcl_wtp failed", __FUNCTION__, __LINE__);
        ifp->if_ierrors++;  // for ifconfig -v
        return NULL;
    }

    return m;
}


/*****************************************************************************/
/*  RX thread it fuction                                                     */
/*****************************************************************************/
void *cpsw_rx_thread (void *arg)
{
    struct cpsw_dev     *cpsw = arg;
    int  rcvid;
    struct _pulse pulse;
    iov_t  msg;

    SETIOV(&msg, &pulse, sizeof(pulse));

    while (1) {
        rcvid = MsgReceivev(cpsw->chid, &msg, 1, NULL);
        if (rcvid == 0) {
            switch (pulse.code)
            {
                case CPSW_RX_PULSE:
                    CpswApp_GetRx(arg);
                break;
                case CPSW_QUIESCE_PULSE:
                    quiesce_block(pulse.value.sival_int);
                break;
                default:
                    slogf(_SLOGC_NETWORK, _SLOG_ERROR,
                        "%s:%d Error: CPSW Rx Unknown pulse %d received", __FUNCTION__, __LINE__, pulse.code);
                break;
            }
        }
        else
        {
            slogf(_SLOGC_NETWORK, _SLOG_ERROR, "%s:%d Error: CPSW Rx MsgReceive error", __FUNCTION__, __LINE__);
        }
    }
    return NULL;
}

